/**************************************************************************/
/*                                                                        */
/*  This file is part of Frama-C.                                         */
/*                                                                        */
/*  Copyright (C) 2007                                                    */
/*    CEA   (Commissariat  l'nergie Atomique)                           */
/*    INRIA (Institut National de Recherche en Informatique et en         */
/*           Automatique)                                                 */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 2.1.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version v2.1                */
/*  for more details (enclosed in the file licenses/LGPLv2.1).            */
/*                                                                        */
/**************************************************************************/

/* Grammar for C annotations */

%{

  open Cil
  open Cil_types
  open Logic_ptree
  open Logic_const

  let loc () = (symbol_start_pos (), symbol_end_pos ())
  let loc_i i = (rhs_start i, rhs_end i)
  let info x = { lexpr_node = x; lexpr_loc = loc () }
  let loc_info loc x = { lexpr_node = x; lexpr_loc = loc }
  let loc_start x = fst x.lexpr_loc
  let loc_end x = snd x.lexpr_loc

  type sense_of_relation = Unknown | Equal | Disequal | Less | Greater

  let relation_sense rel sense =
    match rel, sense with
        Eq, _ -> sense, true
      | Neq, Unknown -> Disequal, true (* No chain of disequality for now*)
      | (Gt|Ge), (Unknown|Equal|Greater) -> Greater, true
      | (Lt|Le), (Unknown|Equal|Less) -> Less, true
      | _ -> sense, false

%}

%token <string> IDENTIFIER STRING_LITERAL TYPENAME
%token <Logic_ptree.constant> CONSTANT
%token LPAR RPAR IF ELSE COLON COLON2 COLONCOLON DOT DOTDOT
%token INT INTEGER REAL FLOAT LT GT LE GE EQ NE COMMA ARROW EQUAL
%token FORALL EXISTS IFF IMPLIES AND OR NOT
%token TRUE FALSE OLD AT RESULT BLOCK_LENGTH BASE_ADDR
%token VALID VALID_INDEX VALID_RANGE FRESH DOLLAR
%token QUESTION MINUS PLUS STAR AMP SLASH PERCENT LSQUARE RSQUARE EOF
%token DATA INVARIANT VARIANT DECREASES FOR LABEL ASSERT SEMICOLON NULL EMPTY
%token REQUIRES ENSURES ASSIGNS LOOP NOTHING SLICE PRAGMA FROM
%token READS LOGIC PREDICATE AXIOM LEMMA LBRACE RBRACE GHOST
%token VOID CHAR SIGNED UNSIGNED SHORT LONG DOUBLE STRUCT ENUM UNION
%token BSUNION INTER
%token LTCOLON COLONGT TYPE BEHAVIOR ASSUMES
%token HAT HATHAT PIPE TILDE GTGT LTLT

%right prec_named
%nonassoc IDENTIFIER
%nonassoc prec_forall prec_exists
%right IMPLIES IFF
%left OR
%left HATHAT
%left AND
%nonassoc prec_not
%nonassoc prec_if
%right QUESTION prec_question
%left PIPE
%left HAT
%nonassoc prec_rel_list
%nonassoc prec_relation
%left LT GT LE GE EQ NE
%left LTLT GTGT
%left PLUS MINUS
%left STAR SLASH PERCENT AMP
%right prec_uminus
%right prec_cast
%nonassoc prec_range
%nonassoc TILDE
%nonassoc LTCOLON COLONGT
%left DOT ARROW LSQUARE
%right prec_par

%type <Logic_ptree.lexpr> lexpr
%start lexpr

%type <Logic_ptree.annot> annot
%start annot

%%

lexpr:
| lexpr_no_rel %prec prec_rel_list { $1 }
| lexpr rel_list %prec prec_rel_list
      { let rel, rhs, _, oth_rel = $2 in
        let loc = loc_start $1, loc_end rhs in
        let relation = loc_info loc (PLrel($1,rel,rhs)) in
        match oth_rel with
            None -> relation
          | Some oth_relation -> info (PLand(relation,oth_relation))
      }

lexpr_dotdot:
| lexpr { $1 }
| lexpr_option DOTDOT lexpr_option { info (PLrange($1,$3)) }

lexpr_no_rel:
  /* predicates */
  lexpr IMPLIES lexpr { info (PLimplies ($1, $3)) }
| lexpr IFF lexpr { info (PLiff ($1, $3)) }
| lexpr OR lexpr     { info (PLor ($1, $3)) }
| lexpr AND lexpr    { info (PLand ($1, $3)) }
| lexpr HATHAT lexpr    { info (PLxor ($1, $3)) }
| NOT lexpr %prec prec_not { info (PLnot $2) }
| TRUE { info PLtrue }
| FALSE { info PLfalse }
| FORALL ne_parameters SEMICOLON lexpr %prec prec_forall
      { info (PLforall ($2, $4)) }
| EXISTS ne_parameters SEMICOLON lexpr %prec prec_exists
      { info (PLexists ($2, $4)) }
| VALID LPAR lexpr RPAR { info (PLvalid ($3)) }
| VALID_INDEX LPAR lexpr COMMA lexpr RPAR { info (PLvalid_index ($3,$5)) }
| VALID_RANGE LPAR lexpr COMMA lexpr COMMA lexpr RPAR
      { info (PLvalid_range ($3,$5,$7)) }
| FRESH LPAR lexpr RPAR { info (PLfresh ($3)) }
| lexpr LTCOLON logic_type_not_id { info (PLinstance_of ($1, $3)) }
| lexpr LTCOLON lexpr { info (PLinstance_ofE ($1, $3)) }
/* terms */
| NULL { info PLnull }
| CONSTANT { info (PLconstant $1) }
| lexpr PLUS lexpr { info (PLbinop ($1, Badd, $3)) }
| lexpr MINUS lexpr { info (PLbinop ($1, Bsub, $3)) }
| lexpr STAR lexpr { info (PLbinop ($1, Bmul, $3)) }
| lexpr SLASH lexpr { info (PLbinop ($1, Bdiv, $3)) }
| lexpr PERCENT lexpr { info (PLbinop ($1, Bmod, $3)) }
| lexpr AMP lexpr { info (PLbinop ($1, Bbw_and, $3)) }
| lexpr PIPE lexpr { info (PLbinop ($1, Bbw_or, $3)) }
| lexpr HAT lexpr { info (PLbinop ($1, Bbw_xor, $3)) }
| lexpr GTGT lexpr { info (PLbinop ($1, Brshift, $3))}
| lexpr LTLT lexpr { info (PLbinop ($1, Blshift, $3))}
| lexpr ARROW identifier { info (PLarrow ($1, $3)) }
| lexpr DOT identifier { info (PLdot ($1, $3)) }
| lexpr LSQUARE lexpr_dotdot RSQUARE { info (PLarrget ($1, $3)) }
| MINUS lexpr %prec prec_uminus { info (PLunop (Uminus, $2)) }
| PLUS lexpr %prec prec_uminus { $2 }
| TILDE lexpr { info (PLunop (Ubw_not, $2)) }
| STAR lexpr { info (PLunop (Ustar, $2)) }
| AMP lexpr { info (PLunop (Uamp, $2)) }
| lexpr QUESTION lexpr COLON2 lexpr %prec prec_question
    { info (PLif ($1, $3, $5)) }
| OLD LPAR lexpr RPAR { info (PLold $3) }
| AT LPAR lexpr COMMA label RPAR { info (PLat ($3, $5)) }
| BASE_ADDR LPAR lexpr RPAR { info (PLbase_addr $3) }
| BLOCK_LENGTH LPAR lexpr RPAR { info (PLblock_length $3) }
| RESULT { info PLresult }
| lexpr COLONGT logic_type_not_id { info (PLcoercion ($1, $3)) }
| lexpr COLONGT lexpr { info (PLcoercionE ($1, $3)) }
/* both terms and predicates */
| identifier LPAR lexpr_list RPAR
    { info (PLapp ($1, [], $3)) }
| identifier LBRACE ne_tvar_list RBRACE LPAR lexpr_list RPAR
    { info (PLapp ($1, $3, $6)) }
| identifier COLON lexpr %prec prec_named { info (PLnamed ($1, $3)) }
| identifier %prec IDENTIFIER { info (PLvar $1) }
| LPAR lexpr_dotdot RPAR %prec prec_par { $2 }
/* Cast. TODO: (identifier *) lexpr needs TYPENAME (see below) */
| LPAR logic_type_not_id RPAR lexpr %prec prec_cast { info (PLcast ($2, $4)) }
| LPAR lexpr_dotdot RPAR lexpr %prec prec_cast
    { match $2.lexpr_node with
	| PLvar x -> info (PLcast (LTnamed (x,[]), $4))
	| _ -> raise Parse_error }
/* tsets */
| EMPTY { info PLempty }
| UNION LPAR lexpr_list RPAR { info (PLunion $3) }
| INTER LPAR lexpr_list RPAR { info (PLinter $3) }
| LBRACE lexpr PIPE ne_parameters RBRACE
      {info (PLcomprehension ($2,$4,None)) }
| LBRACE lexpr PIPE ne_parameters SEMICOLON lexpr RBRACE
      { info (PLcomprehension ($2,$4,Some $6)) }
;

rel_list:
  relation lexpr %prec prec_relation
  { $1, $2, fst(relation_sense $1 Unknown), None }
| relation lexpr_no_rel rel_list %prec prec_relation
  {
    let next_rel, rhs, sense, oth_rel = $3 in
    let (sense, correct) = relation_sense $1 sense
    in
    if correct then
      let loc = loc_start $2, loc_end rhs in
      let my_rel = loc_info loc (PLrel($2,next_rel,rhs)) in
      let oth_rel = match oth_rel with
          None -> my_rel
        | Some rel ->
	    let loc = loc_start $2, loc_end rel in
	    loc_info loc (PLand(rel,my_rel))
      in
      $1,$2,sense,Some oth_rel
    else begin
      let loc = Parsing.rhs_start_pos 1, Parsing.rhs_end_pos 3 in
      raise (Not_well_formed
               (loc,"Inconsistent inequality chain."));

    end
  }
;

lexpr_option:
| /* epsilon */ { None }
| lexpr         { Some $1 }
;

logic_type:
  identifier { LTnamed ($1,[]) }
| identifier stars { $2 (LTnamed ($1,[])) }
| identifier LT ne_logic_type_list GT { LTnamed($1,$3) }
| INTEGER { LTinteger }
| REAL { LTreal }
| logic_type_not_id { $1 }
;

ne_logic_type_list:
  logic_type { [$1] }
| logic_type COMMA ne_logic_type_list { $1 :: $3 }
;

logic_type_not_id:
| VOID           { LTvoid }
| CHAR           { LTint IChar }       /** [char] */
| SIGNED CHAR    { LTint ISChar }      /** [signed char] */
| UNSIGNED CHAR  { LTint IUChar }      /** [unsigned char] */
| SIGNED INT     { LTint IInt }        /** [int] */
| INT            { LTint IInt }        /** [int] */
| UNSIGNED INT   { LTint IUInt }       /** [unsigned int] */
| SIGNED SHORT   { LTint IShort }      /** [short] */
| SHORT          { LTint IShort }      /** [short] */
| UNSIGNED SHORT { LTint IUShort }     /** [unsigned short] */
| SIGNED LONG    { LTint ILong }       /** [long] */
| LONG           { LTint ILong }       /** [long] */
| UNSIGNED LONG  { LTint IULong }      /** [unsigned long] */
| SIGNED LONG LONG   { LTint ILongLong }   /** [long long] (or [_int64] on
					   Microsoft Visual C) */
| LONG LONG   { LTint ILongLong }   /** [long long] (or [_int64] on
					   Microsoft Visual C) */
| UNSIGNED LONG LONG { LTint IULongLong }  /** [unsigned long long]
                                (or [unsigned _int64] on Microsoft Visual C) */
| FLOAT       { LTfloat FFloat }
| DOUBLE      { LTfloat FDouble }
| LONG DOUBLE { LTfloat FLongDouble }
| STRUCT identifier { LTstruct $2 }
| ENUM identifier { LTenum $2 }
| UNION identifier { LTunion $2 }
| TYPENAME         { LTnamed ($1,[]) } /* TODO: Logic_lexer should make it */
| logic_type_not_id STAR { LTpointer $1 }
;

stars:
  STAR { fun t -> LTpointer t }
| stars STAR { fun t -> $1 (LTpointer t) }
;

relation:
  | LT    { Lt }
  | GT    { Gt }
  | LE    { Le }
  | GE    { Ge }
  | EQ    { Eq }
  | NE    { Neq }
;

lexpr_list:
| /* epsilon */ { [] }
| ne_lexpr_list  { $1 }
;

ne_lexpr_list:
| lexpr                    { [$1] }
| lexpr COMMA ne_lexpr_list { $1 :: $3 }
;

requires:
  /* epsilon */ { [] }
| REQUIRES lexpr SEMICOLON requires { $2::$4 }
;


behavior_body:
  /* epsilon */ { [],[],[] }
| ne_behavior_body { $1 }

ne_behavior_body:
| ASSUMES lexpr SEMICOLON behavior_body
    { let a,b,c = $4 in $2::a,b,c }
| ENSURES lexpr SEMICOLON behavior_body
    { let a,b,c = $4 in a,$2::b,c }
| ASSIGNS location_dep SEMICOLON behavior_body
    { let a,b,c = $4 in a,b,$2@c }
| ASSIGNS NOTHING SEMICOLON behavior_body
    { let a,b,c = $4 in
      match c with
          [] | [(Nothing,_)] -> a,b,[(Nothing,[])]
        | _ -> raise (
            Not_well_formed (loc(),"Mixing \\nothing and a real location"))
    }
;

behaviors:
  /* epsilon */ { [] }
| BEHAVIOR identifier COLON behavior_body behaviors
      { let (assumes,ensures,assigns) = $4 in
        Logic_const.check_assigns ~loc:(loc()) assigns;
        {b_name=$2; b_assumes = assumes;
         b_ensures = ensures;
         b_assigns = assigns}::$5 }

behaviors_or_default:
  behaviors { $1 }
| ne_behavior_body { let (assumes,ensures,assigns) = $1 in
                     [{ b_name = "default";b_assumes = assumes;
                        b_ensures = ensures;
                        b_assigns = assigns} ]}
;

spec:
  requires behaviors_or_default decreases
    { { spec_requires = $1; spec_behavior = $2; spec_variant = $3 } }
;

loop_annot:
  loop_invariant { [AInvariant (true,$1)] }
| loop_effects { List.map (fun x -> AAssigns x) $1 }
| loop_variant { [AVariant $1] }
| loop_pragma { [APragma (Loop_pragma $1)] }
;

loop_annotations:
  loop_annot
    { $1 }
| loop_annot loop_annotations { $1 @ $2 }
;

type_annot:
| TYPE INVARIANT identifier LPAR logic_type identifier RPAR EQUAL lexpr SEMICOLON
  { { inv_name = $3; this_name = $6; this_type = $5; inv = $9; } }
;

variant:
  lexpr FOR identifier { ($1, Some $3) }
| lexpr                { ($1, None) }
;

loop_invariant:
  LOOP INVARIANT lexpr SEMICOLON { $3 }
;

loop_variant:
  LOOP VARIANT variant SEMICOLON { $3 }
;

decreases:
  /* epsilon */   { None }
| DECREASES variant SEMICOLON { Some $2 }
;

code_annotation:
  slice_pragma     { APragma (Slice_pragma $1) }
| ASSERT lexpr SEMICOLON    { AAssert $2 }
| INVARIANT lexpr SEMICOLON { AInvariant (false,$2) }

loop_pragma_tk:
  LOOP PRAGMA { }
;

loop_pragma:
  loop_pragma_tk IDENTIFIER ne_lexpr_list SEMICOLON
  { if $2 = "UNROLL_LOOP" || $2 = "UNROLL" then
      match $3 with
        | [level] -> Unroll_level level
        | _ -> raise(
            Not_well_formed(loc(),"usage: loop pragma unroll n;"))
    else if $2 = "WIDEN_VARIABLES" then
      Widen_variables $3
    else if $2 = "WIDEN_HINTS" then
      Widen_hints $3
    else raise (Not_well_formed (loc(),"unknown loop pragma")) }

slice_pragma_tk:
  SLICE PRAGMA { }
;

slice_pragma:
  slice_pragma_tk IDENTIFIER lexpr SEMICOLON
    { if $2 = "expr" then SPexpr $3
      else raise (Not_well_formed (loc(), "unknown slice pragma")) }
| slice_pragma_tk IDENTIFIER SEMICOLON
    { if $2 = "ctrl" then SPctrl
      else if $2 = "stmt" then SPstmt
      else raise (Not_well_formed (loc(), "unknown slice pragma")) }

loop_effects:
  LOOP ASSIGNS location_dep SEMICOLON { $3 }
| LOOP ASSIGNS NOTHING SEMICOLON { [(Nothing,[])] }
;

location_dep:
| zones { List.map (fun x -> (x,[])) $1 }
| zones FROM zones {List.map (fun x -> (x, $3)) $1}
| zones FROM NOTHING  {List.map (fun x -> (x, [Nothing])) $1}

zones : tsets { List.map (fun x -> Location x) $1 }

tsets:
| location { [$1] }
| location COMMA tsets { $1 :: $3 }
;

location:
| lexpr { $1 }
;

logic_decl:
| LOGIC logic_type poly_id LPAR parameters RPAR
  { let (id,labels,tvars) = $3 in
    ($2,id,labels, tvars,$5) }

poly_id:
| identifier { ($1,[],[]) }
| identifier LT ne_tvar_list GT { $1,[],$3 }
| identifier LBRACE ne_label_list RBRACE { ($1,$3,[]) }
| identifier LBRACE ne_label_list RBRACE LT ne_tvar_list GT { $1,$3,$6 }

identifier:
| IDENTIFIER { $1 }
| ASSERT { "assert" }
| ASSIGNS { "assigns" }
| ASSUMES { "assumes" }
| AXIOM { "axiom" }
| BEHAVIOR { "behavior" }
| DATA { "data" }
| DECREASES { "decreases" }
| ENSURES  { "ensures" }
/* | INTEGER { "integer" } */
| INVARIANT { "invariant" }
| LABEL { "label" }
| LEMMA { "lemma" }
| LOGIC { "logic" }
| LOOP { "loop" }
| PRAGMA { "pragma" }
| PREDICATE { "predicate" }
| READS { "reads" }
/* | REAL { "real" } */
| REQUIRES { "requires" }
| SLICE { "slice" }
| TYPE { "type" }
| VARIANT { "variant" }
;

decl:
| logic_decl SEMICOLON
    { let (rt, id, labels, tvars, args) = $1 in
      LDlogic_reads (id, labels, tvars, rt, args, []) }
| logic_decl READS tsets SEMICOLON
    { let (rt, id, labels, tvars, args) = $1 in
      LDlogic_reads (id, labels, tvars, rt, args, $3) }
| logic_decl EQUAL lexpr SEMICOLON
    { let (rt, id, labels, tvars, args) = $1 in
      LDlogic_def (id, labels, tvars, rt, args, $3) }
| TYPE poly_id SEMICOLON
    { let (id,labels,tvars) = $2 in
      assert (labels = []);
      LDtype(id,tvars) }
| PREDICATE poly_id LPAR parameters RPAR SEMICOLON
    { let (id,labels,tvars) = $2 in
      LDpredicate_reads (id, labels, tvars, $4, []) }
| PREDICATE poly_id LPAR parameters RPAR READS tsets SEMICOLON
    { let (id,labels,tvars) = $2 in
      LDpredicate_reads (id, labels, tvars, $4, $7) }
| PREDICATE poly_id LPAR parameters RPAR EQUAL lexpr SEMICOLON
    { let (id,labels,tvars) = $2 in
      LDpredicate_def (id, labels, tvars, $4, $7) }
| AXIOM poly_id COLON lexpr SEMICOLON
    { let (id,labels,tvars) = $2 in
      LDlemma (id, true, labels, tvars, $4) }
| LEMMA poly_id COLON lexpr SEMICOLON
    { let (id,labels,tvars) = $2 in
      LDlemma (id, false, labels, tvars, $4) }
| DATA INVARIANT identifier COLON lexpr SEMICOLON
    { LDinvariant ($3, $5) }
| type_annot {LDtype_annot $1}
;

ne_tvar_list:
| identifier { [$1] }
| identifier COMMA ne_tvar_list { $1 :: $3 }
;

ne_label_list:
| label { [$1] }
| label COMMA ne_label_list { $1 :: $3 }
;

label:
| identifier { $1 }
;

parameters:
| /* epsilon */  { [] }
| ne_parameters { $1 }
;

ne_parameters:
  parameter { [$1] }
| parameter COMMA ne_parameters { $1 :: $3 }
;

parameter:
  logic_type identifier { ($1, $2) }
| logic_type identifier LSQUARE RSQUARE { (LTarray $1, $2) }
;

annot:
  annotation EOF   { $1 }
;

annotation:
  decl             { Adecl (loc (), $1) }
| spec             { Aspec (loc (), $1) }
| loop_annotations { Logic_const.check_loop_annotation ~loc:(loc()) $1;
                     Aloop_annot (loc (), $1) }
| code_annotation { Acode_annot (loc(),$1) }
| IDENTIFIER      { Aattribute_annot (loc (), $1) }

;

%%
