(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007                                                    *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(** Main GUI skeleton *)
open Cil_types
open Cil
open Db_types
open Db
open Pretty_source
open Gtk_helper

let run_configure_dialog () = 
  let d = GWindow.dialog ~position:`CENTER ~modal:true ~title:"Configure analysis" ~show:true () 
  in
  d#add_button_stock `CLOSE `CLOSE;
  d#set_default_response `CLOSE;
  let on_bool label get set = 
    let button = GButton.check_button 
      ~label
      ~packing:d#vbox#add
      ~active:(get ())
      ()
    in
    ignore (button#connect#toggled ~callback:(fun () -> set button#active))
  in
  on_bool "Values" Cmdline.ForceValues.get Cmdline.ForceValues.set;
  on_bool "Functional dependencies" 
    Cmdline.ForceDeps.get Cmdline.ForceDeps.set;
  on_bool "Call-site wise functional dependencies " 
    Cmdline.ForceCallDeps.get Cmdline.ForceCallDeps.set;
(*  Cmdline.iter_on_options 
    (fun (k, s) -> match k with
     | Cmdline.Bool (get, set) -> on_bool s get set
     | Cmdline.Int _ -> ());*)
  match d#run () with 
  | `DELETE_EVENT | `CLOSE -> d#destroy ()
    
let use_external_viewer = false

let highlight_range ~scroll tag (v:GText.view) pb pe =
  let b = v#buffer in
  let start = b#get_iter (`OFFSET pb) in
  let stop = b#get_iter (`OFFSET pe) in
  b#apply_tag tag start stop;
  if scroll then begin
    b#place_cursor start;
    ignore (v#scroll_to_mark `INSERT)
  end

(** The initial menus and toolbar *)
let initial_ui_info = "<ui>
  <menubar name='MenuBar'>
    <menu action='FileMenu'>
      <separator/>
    </menu>
    <menu action='AnalysesMenu'>
      <separator/>
    </menu>
  </menubar>
  <toolbar name='ToolBar'>
  </toolbar>
</ui>"

let basic_ui_info = "<ui>
  <menubar name='MenuBar'>
    <menu action='FileMenu'>
      <menuitem action='Load'/>
      <menuitem action='Save'/>
      <menuitem action='Quit'/>
    </menu>
    <menu action='AnalysesMenu'>
      <menuitem action='Configure'/>
      <menuitem action='Run'/>
    </menu>
  </menubar>

  <toolbar name='ToolBar'>
      <toolitem action='Load'/>
      <toolitem action='Save'/>
      <toolitem action='Run'/>
      <toolitem action='Configure'/>
      <toolitem action='Quit'/>
  </toolbar>
</ui>"

(** *)
class type main_window_extension_points = object
  method ui_manager : GAction.ui_manager
  method actions : GAction.action_group

  method file_tree : Filetree.t
  method toplevel : main_window_extension_points
  method main_window : GWindow.window
  method annot_window : GText.view

  method source_viewer : GText.view
  method display_globals : global list -> GText.buffer -> unit
  method register_source_selector : (GMenu.menu GMenu.factory -> main_window_extension_points -> button:int -> Pretty_source.localizable -> unit) -> unit

  method register_source_highlighter : (GText.buffer -> localizable -> start:int -> stop:int -> unit) -> unit

  method highlight : scroll:bool -> GText.tag -> Pretty_source.localizable -> unit

  method original_source_viewer : Source_manager.t
  method view_original_stmt : stmt -> location
  method view_original : location -> unit

  method reset : unit -> unit

  method monospace : Pango.font_description
  method general : Pango.font_description

  method info : 'a. ('a, Format.formatter, unit) format -> 'a
end
    
(** Load a project file *)
let load_file (main_ui:main_window_extension_points) =
  let dialog = GWindow.file_chooser_dialog
    ~action:`OPEN
    ~title:"Open state"
    ~parent:main_ui#main_window () in
  dialog#add_button_stock `CANCEL `CANCEL ;
  dialog#add_select_button_stock `OPEN `OPEN ;
  begin match dialog#run () with
  | `OPEN ->
      begin match dialog#filename with
      | None -> ()
      | Some f ->
          (try Project.load_all f
            with Sys_error s -> main_ui#info "Cannot load: %s" s);
          main_ui#reset ()
      end
  | `DELETE_EVENT | `CANCEL -> ()
  end ;
  dialog#destroy ()

(** Load a project file *)
let save_file (main_ui:main_window_extension_points) =
  let dialog = GWindow.file_chooser_dialog
    ~action:`SAVE
    ~title:"Save state"
    ~parent:main_ui#main_window () in
  dialog#add_button_stock `CANCEL `CANCEL ;
  dialog#add_select_button_stock `SAVE `SAVE ;
  begin match dialog#run () with
  | `SAVE ->
      begin match dialog#filename with
      | None -> ()
      | Some f ->
          try Project.save_all f
          with Sys_error s -> main_ui#info "Cannot save: %s" s
      end
  | `DELETE_EVENT | `CANCEL -> ()
  end ;
  dialog#destroy ()

(** The list of registered extension *)
let (handlers:(main_window_extension_points -> unit) list ref) = ref []
(** Insert an extension *)
let register_extension f =
  handlers := f::!handlers
(** Apply all extensions *)
let process_extensions window =
  List.iter (fun f -> f window) (List.rev !handlers)

(** The list of reset extensions. Such extension are used for example when the current project
    is changed. *)
let (reset_handlers:(main_window_extension_points -> unit) list ref) = ref []
(** Insert a reset extension *)
let register_reset_extension f =
  reset_handlers := f::!reset_handlers
(** Apply all reset extensions *)
let reset_extensions window =
  List.iter (fun f -> f window) (List.rev !reset_handlers)

let filetree_selector (main_ui:main_window_extension_points) deselecting globals =
  let source = main_ui#source_viewer in
  if deselecting then true (* unselecting this one *)
  else try
    (*Format.printf "ROW ACTIVATED %s@." path_s;*)
    let buffer = Source_viewer.buffer () in

    (* Use a fresh buffer: this prevents source code from being redrawn before
       it is completly written --> speed is higher.*)
    main_ui#display_globals globals buffer;
    source#set_buffer buffer;
    true
  with e ->
    main_ui#info "gui could not select row, please report: %s"
      (Printexc.to_string e);
    false

let rec to_do_on_select
    (menu_factory:GMenu.menu GMenu.factory)
    (main_ui:main_window_extension_points)
    ~button
    selected
    =
  let tree_view = main_ui#file_tree in
  let annot = main_ui#annot_window#buffer in
  if button = 1 then
    begin match selected with
    | PStmt (kf,ki) -> begin
        (* Find kinstr and kf *)
	try
          let loc = main_ui#view_original_stmt ki in
          let skind =
            if (Cmdline.Debug.get ()> 0) then
              match ki with
              | {skind=Block _} -> "Block "
              | {skind=Instr (Skip _)} -> "Skip "
              | _ -> ""
            else ""
          in
	  let n =
            Format.sprintf
              "Function: %s@\n%sStatement: %d (line %d)@\n"
              (fprintf_to_string "%a" Kernel_function.pretty_name kf)
              skind
              ki.sid
              (fst loc).Lexing.pos_lnum
          in
	  annot#set_text n;

          (* Code annotations for this statement *)
	  List.iter
	    (fun a ->
	       let pos,a = match a with
	       | Before a -> "Before", a
	       | After a -> "After", a
	       in
	       let user,s = match a with
	       | WP (a,_) ->
		   "WP",
		   begin
                     try
                       let a = List.find
                         (function Fol.Goal _ -> true | _ -> false)
                         a
                       in
                       let a = match a with
                       | Fol.Goal (_,a) -> a
                       | _ -> assert false
                       in
                       fprintf_to_string "%a" Why_output.predicate a
                     with Not_found -> "no such goal"
                   end
	       | User a ->
		   "user",
		   fprintf_to_string "%a" !Ast_printer.d_code_annotation a
	       | AI (_,a) ->
                   "alarm",
		   fprintf_to_string "%a" !Ast_printer.d_code_annotation a
	       in
	       annot#insert (Format.sprintf "%s(%s): %s@\n" pos user s))
	    (Annotations.get ki);

	with e ->
	  annot#insert ("error in Db.KernelFunction.find ("
                        ^ (Printexc.to_string e)
                        ^ ")")
      end
    | PLval (kf, ki,lv) ->
        begin try
          begin
            match ki with Kglobal -> ()
            | Kstmt ki -> ignore (main_ui#view_original_stmt ki)
          end;
          let ty = typeOfLval lv in
          if isFunctionType ty then
            annot#set_text "This is a C function\n"
          else begin
            let sid = Ast_info.get_sid ki in
	    let n =
              Format.sprintf
                "Function: %s@\nStatement: %d@\n"
                (match kf with None -> "<none>"
                 | Some kf ->
		     fprintf_to_string "%a" Kernel_function.pretty_name kf)
                sid
            in
	    annot#set_text n;
            let vars = extract_varinfos_from_lval lv in
            VarinfoSet.iter
              (fun vi ->
                 annot#insert
                   (Format.sprintf
                      "Variable %s has type \"%s\".\nIt is a %s variable.\n%sIt is %sreferenced and its address is %staken.\n"
                      (fprintf_to_string "%a" Ast_info.pretty_vname vi)
                      (fprintf_to_string "%a" !Ast_printer.d_type vi.vtype)
                      (if vi.vglob then "global" else "local")
                      (match vi.vdescr with None -> ""
                       | Some s ->
                           Format.sprintf "This is a temporary variable for \"%s\".\n" s)
                      (if vi.vreferenced then "" else "not ")
                      (if vi.vaddrof then "" else "not ")))
              vars
          end
        with Not_found ->
          annot#insert "Error in lval Db.KernelFunction.find\n"
        end
    | PTermLval _ ->
	() (* JS: TODO (?) *)
    | PVDecl (kf,vi) ->
        main_ui#view_original vi.vdecl;
        if vi.vglob
        then
          annot#set_text
            (Format.sprintf "This is the declaration of global %s\nIt is %sreferenced and its address is %staken.\n"
               (fprintf_to_string "%a" Ast_info.pretty_vname vi)
               (if vi.vreferenced then "" else "not ")
               (if vi.vaddrof then "" else "not "))
        else
          annot#set_text
            (Format.sprintf "This is the declaration of local %s in function %s\n%s"
               (fprintf_to_string "%a" Ast_info.pretty_vname vi)
               (fprintf_to_string "%a"
		  Kernel_function.pretty_name (Cilutil.out_some kf))
               (match vi.vdescr with None -> ""
                | Some s ->  Format.sprintf "This is a temporary variable for \"%s\".\n" s))
    end
  else if button = 3 then begin
    match selected with
    | PVDecl _ -> ()
    | PStmt (kf,ki) ->
        let add_assert_after _ =
          let txt =
            GToolbox.input_string
              ~title:"Insert an assertion after"
              ""
          in
          match txt with
          | None -> ()
          | Some txt -> try
              Db.Properties.add_assert
		kf ki ~before:false
		("assert " ^ txt ^ ";");
              to_do_on_select menu_factory main_ui ~button:1 selected
	    with e ->
	      annot#set_text (Printexc.to_string e)
        in
        let add_assert_before _ =
          let txt =
            GToolbox.input_string
              ~title:"Insert an assertion before"
              ""
          in
          match txt with
          | None -> ()
          | Some txt -> try
              Db.Properties.add_assert
		kf ki ~before:true ("assert " ^ txt ^ ";");
              to_do_on_select menu_factory main_ui ~button:1 selected
	    with e ->
	      annot#set_text (Printexc.to_string e)
        in
        begin
          ignore (menu_factory#add_item
		    "Add asser _after" ~callback:add_assert_after);
          ignore (menu_factory#add_item
		    "Add asser _before" ~callback:add_assert_before)
        end
    | PLval (_kf, _ki, lv) ->
        let ty = typeOfLval lv in
        (* Do special actions for functions *)
        (* popup a menu to jump the definitions of the given varinfos *)
        let do_menu l =
          match l with
          | [] -> ()
          | _ ->
              try
                List.iter
                  (fun v ->
                     ignore
                       (menu_factory#add_item
                          ("Go to definition of " ^
                             (Pretty_utils.escape_underscores
				(fprintf_to_string "%a"
				   Ast_info.pretty_vname v)))
                          ~callback:
                          (fun () ->
                             tree_view#select_global v)))
                  l
              with Not_found -> ()
        in
        (match lv with
         | Var v,NoOffset when isFunctionType ty ->
             (* only simple literal calls can be resolved syntactically *)
             do_menu [v]
         | _  -> ())
    | PTermLval _ ->
	() (* JS: TODO (?) *)

  end

(** The main application window *)
class main_window () : main_window_extension_points =
  let main_window =
    GWindow.window
      ~title:"Frama-C"
      ~width:1024 ~height:768
      ~allow_shrink:true
      ~allow_grow:true
      ()
  in

  (* On top one finds the menubar *)
  let toplevel_vbox = GPack.box `VERTICAL ~packing:main_window#add ()
  in

  (* Build the main menubar *)
  (* The menu thing with one default action to quit *)
  let actions = GAction.action_group ~name:"Actions" () in
  (* Make the ui manager thing *)
  let ui_manager = GAction.ui_manager () in
  (* Connect accelerators *)
  let () = main_window#add_accel_group ui_manager#get_accel_group in

  let () = 
    GAction.add_actions actions 
      [ GAction.add_action "FileMenu" ~label:"_File";
        GAction.add_action "AnalysesMenu" ~label:"_Analyses" ;
      ]
  in
    
  (* Register actions for menubar *)
  let () =
    ui_manager#insert_action_group actions 0;
    ignore (ui_manager#add_ui_from_string initial_ui_info )
  in

  (* And now: the menubar *)
  let menubar = new GMenu.menu_shell
    (GtkMenu.MenuBar.cast (ui_manager#get_widget "ui/MenuBar")#as_widget)
  in
  (* Pack the menu bar on top of gui *)
  let () = toplevel_vbox#pack ~fill:false menubar#coerce
  in

  (* And the toolbar *)
  let toolbar = new GButton.toolbar
    (GtkButton.Toolbar.cast (ui_manager#get_widget "ui/ToolBar")#as_widget)
  in
  (* Pack the menu bar on top of gui *)
  let () = toplevel_vbox#pack ~fill:false toolbar#coerce
  in

  (* Split below the bars *)
  let toplevel_hpaned = GPack.paned `HORIZONTAL ~packing:toplevel_vbox#add () in

  (* The left filetree inside an automatic scrolled window and a nice frame *)
  let filetree_frame = GBin.frame ~shadow_type:`ETCHED_OUT ~packing:toplevel_hpaned#add1 () in
  let filetree_scrolled_window =
    GBin.scrolled_window
      ~vpolicy:`AUTOMATIC
      ~hpolicy:`NEVER
      ~packing:filetree_frame#add ()
  in
  let file_tree = GTree.view ~packing:filetree_scrolled_window#add () in
  let () = file_tree#selection#set_mode `BROWSE in
  let () = file_tree#set_rules_hint true in
  let file_tree = Filetree.make file_tree in

  (* status bar (at bottom) *)
  let statusbar = GMisc.statusbar ~packing:(toplevel_vbox#pack ~fill:false) () in

  let status_context = statusbar#new_context "messages" in
  let _ = status_context#push "Ready" in
  let _ = status_context#push "Ready" in

  (* splits between messages and sources *)
  let vb_message_sources =
    GPack.paned `VERTICAL  ~border_width:3 ~packing:toplevel_hpaned#add2 ()
  in

  (* splits between messages and sources *)
  let hb_sources =
    GPack.paned `HORIZONTAL  ~border_width:3 ~packing:vb_message_sources#add1 ()
  in

  (* Set the relative position for all paned whenever the main window is resized *)
  let () =
    ignore (main_window#misc#connect#size_allocate
	      (let old_w = ref 0
	       and old_h = ref 0 in
	       fun {Gtk.width=w;Gtk.height=h} ->
	         if !old_w <> w or !old_h <> h then
		   begin
		     old_h := h;
		     old_w := w;
                     hb_sources#set_position
                       ((hb_sources#max_position - hb_sources#min_position)/2);
                     vb_message_sources#set_position
                       ((vb_message_sources#max_position - vb_message_sources#min_position)*5/7);
                     (* toplevel_hpaned#set_position
                        ((toplevel_hpaned#max_position - toplevel_hpaned#min_position)*3/10);*)
		   end))
  in

  (* lower notebook *)
  let fr2 = GBin.frame ~shadow_type:`ETCHED_OUT ~packing:vb_message_sources#add2 () in
  let lower_notebook =
    GPack.notebook ~scrollable:true ~show_tabs:true ~packing:fr2#add ()
  in

  (*  let count = ref 0 in
      let _id = Glib.Timeout.add
      ~ms:10000
      ~callback:(fun () ->
      Warning_manager.append
      warning_manager
      { Warning_manager.m_file = "Here";
      m_line = (incr count; !count);
      m_msg = "Watch me!" ;
      m_severity = `Warn}
      ~on_select:(fun s -> assert false);
      true(*repeat ad lib*)
      )
      in
  *)
  (* lower text view and its scroll view: annotations and messages *)
  let sw = GBin.scrolled_window
    ~vpolicy:`AUTOMATIC
    ~hpolicy:`AUTOMATIC
    ~packing:
    (fun w -> ignore (lower_notebook#prepend_page
                        ~tab_label:(GMisc.label ~text:"Informations" ())#coerce w))
    ()
  in

  let annot_window = GText.view ~packing:sw#add () in
  let _ = annot_window#set_editable false in

  (* upper text view: source code *)
  let fr1 = GBin.frame ~shadow_type:`ETCHED_OUT ~packing:hb_sources#add1 () in
  let sw = GBin.scrolled_window
    ~vpolicy:`AUTOMATIC
    ~hpolicy:`AUTOMATIC
    ~packing:fr1#add ()
  in
  let source_viewer = Source_viewer.make ~packing:sw#add in

  let original_source_viewer = Source_manager.make ~packing:hb_sources#add2 in

  (* Remove default pango menu for textviews *)
  let () =
    ignore (source_viewer#event#connect#button_press ~callback:
	      (fun ev -> GdkEvent.Button.button ev = 3));
    ignore (annot_window#event#connect#button_press ~callback:
	      (fun ev -> GdkEvent.Button.button ev = 3));
    (* startup configuration *)
    source_viewer#buffer#place_cursor
      ~where:source_viewer#buffer#start_iter
  in
object (self:#main_window_extension_points)
  val mutable selector = []
  val mutable highlighter = []
    (* Warning list *)

  method monospace = Pango.Font.from_string (Cmdline.MonospaceFontName.get ())
  method general = Pango.Font.from_string (Cmdline.GeneralFontName.get ())
  method toplevel = (self:>main_window_extension_points)
  method main_window = main_window
  method ui_manager = ui_manager
  method actions = actions
  val mutable file_tree = file_tree
  method file_tree = file_tree
  method annot_window = annot_window
  method source_viewer = source_viewer
  method register_source_selector f = selector <- f::selector
  method register_source_highlighter f = highlighter <- f::highlighter
  method original_source_viewer = original_source_viewer
  method display_globals globs buffer =
    Pretty_source.display_source
      globs
      buffer
      ~highlighter:(fun localizable ~start ~stop ->
                      List.iter (fun f -> f buffer localizable ~start ~stop) highlighter)
      ~selector:(fun ~button localizable ->
                   let popup_factory = new GMenu.factory (GMenu.menu ()) in
                   List.iter
                     (fun f ->
                        try
			  f popup_factory
                            self#toplevel ~button
			    localizable
			with e ->
			  self#info
			    "Selector got exception %s; please report."
			    (Printexc.to_string e))
                     selector;
                   if button = 3 && popup_factory#menu#children <> [] then
                     popup_factory#menu#popup
                       ~button
                       ~time:(GtkMain.Main.get_current_event_time ()))

  method highlight ~scroll tag loc =
    match Pretty_source.locate_localizable loc with
    | None ->
        if Cmdline.Debug.get () > 0 then
          self#info "Could not highlight in GUI."
    | Some (b,e) ->
        highlight_range ~scroll tag self#source_viewer b e

  method view_original loc =
    if loc <> locUnknown then
      Source_manager.load_file
        self#monospace
        self#original_source_viewer
        (fst loc).Lexing.pos_fname
        (fst loc).Lexing.pos_lnum

  method view_original_stmt st =
    let loc = get_stmtLoc st.skind in
    if use_external_viewer
    then
      (if loc <> locUnknown then
         let args_for_emacs =
	   Format.sprintf "emacsclient -n +%d %s" 
             (fst loc).Lexing.pos_lnum (fst loc).Lexing.pos_fname
             (*          Format.sprintf "mate -a -l %d %s" line file  *)
         in
         if Cmdline.Debug.get_cmdline () > 0 then
           self#info "Running %s" args_for_emacs;
         ignore (Sys.command args_for_emacs))
    else self#view_original loc;
    loc

  method private info_string s = 
    status_context#pop ();
    ignore (status_context#push s)

  method info fmt =
    let b = Buffer.create 80 in
    let bfmt = Format.formatter_of_buffer b in
    Format.kfprintf
      (function fmt ->
	 Format.pp_print_flush fmt ();
	 let content = Buffer.contents b in
         self#info_string content)
      bfmt 
      fmt

(*  flash_info := !display_info (* fun s -> status_context#flash ~delay:10 s *);*)

  (* These private method might be exported when necessary *)
  method private toplevel_vbox = toplevel_vbox
  method private toplevel_hpaned = toplevel_hpaned
  method private statusbar = statusbar
  method private menubar = menubar
  method private reparent parent =
    toplevel_vbox#misc#reparent parent;
    main_window#destroy ()
  method reset () =
    source_viewer#buffer#set_text "Please select a file in the left panel.";
    file_tree <- self#file_tree#reset ();
    reset_extensions self#toplevel
  initializer
    
    main_window#misc#modify_font self#general;
    ignore (main_window#connect#destroy ~callback:(fun () -> exit 0));
    self#annot_window#misc#modify_font self#general;
    Gtk_helper.redirect Format.std_formatter self#annot_window#buffer;
    Gtk_helper.redirect Format.err_formatter self#annot_window#buffer;
    statusbar#misc#modify_font self#general;
    file_tree#set_select_function (filetree_selector self#toplevel);
    process_extensions self#toplevel;
    self#register_source_selector to_do_on_select;
    let () = GAction.add_actions actions
      [ 
        GAction.add_action "Load" ~stock:`OPEN ~tooltip:"Load a saved state"
          ~accel:"<control>O"
	  ~callback:(fun _ -> load_file self#toplevel) ;
        GAction.add_action "Save" ~stock:`SAVE ~tooltip:"Save state"
          ~accel:"<control>S"
	  ~callback:(fun _ -> save_file self#toplevel) ;
        
        GAction.add_action "Configure" ~stock:`PREFERENCES ~tooltip:"Configure analyses"
	  ~callback:(fun _ -> run_configure_dialog ());
        GAction.add_action "Run" ~stock:`EXECUTE ~tooltip:"Run analyses"
          ~accel:"<control>R"
	  ~callback:(fun _ ->
                       (*let old = Project.current () in
                       let project = Project.create "interactive" in
                       Project.set_current project;
			 Project.copy ~src:old ~only:(Cmdline.get_selection ()) project;*)
		       annot_window#buffer#set_text "Running analysis\n";
                       !Db.Toplevel.run_all_plugins 
			 (Gtk_helper.make_formatter annot_window#buffer);
                       self#reset ()) ;
        GAction.add_action "Quit" ~stock:`QUIT ~tooltip:"Quit" ~accel:"<control>Q"
	  ~callback:(fun _ -> exit 0);
      ]
    in
    ignore (ui_manager#add_ui_from_string basic_ui_info);

    main_window#show ();
    let warning_manager =
      Warning_manager.make
        ~packing:(fun w ->
                    ignore (lower_notebook#append_page
                              ~tab_label:(GMisc.label ~text:"Messages" ())#coerce w))
        ~callback:(fun s d ->
                     let locs = localizable_from_locs s d in
                     match locs with
                     | [] -> 
			 let loc = { Lexing.dummy_pos with 
			               Lexing.pos_lnum=d; Lexing.pos_fname=s;
			           } in
			 self#view_original (loc,loc)
                     | loc::_ ->
                         to_do_on_select
                           (new GMenu.factory (GMenu.menu ()))
                           ~button:1
                           self#toplevel
                           loc)
    in

    Hashtbl.iter
      (fun _k mess ->
         Warning_manager.append
           warning_manager
           mess
           ~on_select:(fun _ -> assert false))
      Messages_manager.warn_table;

end

let toplevel_init () =
  Cilutil.flush_all ();
  ignore (GtkMain.Main.init ());
  ignore (new main_window ());
  GMain.Main.main ()

let () = Messages_manager.enable_collect ()

let () = Options.add_plugin
  ~name:"ValViewer"
  ~descr:"gui to display values"
  ~toplevel_init
  ["-monospace-font",
   Arg.String Cmdline.MonospaceFontName.set,
   Format.sprintf "f : use font f as monospace font (defaults to %s)"
     (Cmdline.MonospaceFontName.get ());
   "-general-font",
   Arg.String Cmdline.GeneralFontName.set,
   Format.sprintf "f : use font f as general font (defaults to %s)"
     (Cmdline.GeneralFontName.get ())
  ]

(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.. -j"
End:
*)
