(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007                                                    *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* $Id: cmdline.ml,v 1.144 2008/03/06 14:10:17 uid528 Exp $ *)

(** Bunch of values which may be initialized through command line. *)

(* ****************************************************************************
   Note to the developper: how to add a new option ?
   =================================================

   * If the options is shared by all projects:
   simply implement the signature [S] and use a reference for the option

   * If the option may to not be shared by all projects:
   - if its datatype is simple:
   simply apply one of the predefined functors (e.g. [False]): the option is a
   module implementing the signature [S].
   - if its datatype is not so simple:
   have a look at the predefined functor IndexedVal, which uses strings as
   indexes for the valid options, and stores the current string index in
   the project's state. Even if you need more than that, this should be an
   example on how to use  [Project.State.Register] to have your options
   properly registered in the current project.
   ************************************************************************* *)

let selection = ref Project.Selection.empty
let get_selection () = !selection
let add_option select =
  selection := select Kind.Do_Not_Select_Dependencies !selection

let nb_selected_options_ref = ref 0
let nb_selected_options () = !nb_selected_options_ref

(* ************************************************************************* *)
(** {2 Generic functors} *)
(* ************************************************************************* *)

module type S = sig
  type t
  val set: t -> unit
  val get: unit -> t
  val get_cmdline: unit -> t
  val clear: unit -> unit
  val is_set: unit -> bool
  include Project.Computation.OUTPUT
end

type kind =
  | Bool of (unit -> bool) * (bool -> unit)
  | Int of (unit -> int) * (int -> unit)

type t = kind * string

let options : t list ref = ref []
let iter_on_options f = List.iter f !options

module Build(X:sig
	       type t
	       val default: t
	       val name: string
	       val equal: t -> t -> bool
	     end)
   : S with type t = X.t
= struct

  let cmd_state = ref X.default
  let on = ref false

  module D = struct
    include Project.Datatype.Persistent(X)
    let default = X.default
    let equal = X.equal
  end
  include Computation.Ref
    (D)
    (struct
       let name = Project.Computation.Name.make X.name
       let dependencies = []
     end)

  type t = data

  let () = add_option select

  let set x =
    on := true;
    set x;
    cmd_state := x;
    if not (X.equal X.default x) then incr nb_selected_options_ref

  let clear () = set X.default

  let get_cmdline () = !cmd_state
  let is_set () = !on

end

(** {3 Boolean} *)

module type BOOL = sig
  include S with type t = bool
  val on: unit -> unit
  val off: unit -> unit
end

module Bool(X:sig val default: bool val name: string end) = struct
  include Build(struct type t = bool include X let equal = (=) end)
  let on () = set true
  let off () = set false
  let () =
    options := (Bool(get, set), Project.Computation.Name.get name) :: !options
end

(** Build a boolean option initialized to [false]. *)
module False(X: sig val name: string end) =
  Bool(struct include X let default = false end)

(** Build a boolean option initialized to [true]. *)
module True(X: sig val name: string end) =
  Bool(struct include X let default = true end)

(** {3 Integer} *)

module type INT = sig
  include S with type t = int
  val incr: unit -> unit
end

(** Build an integer option. *)
module Int(X: sig val default: int val name: string end) = struct
  include Build(struct type t = int include X let equal = (=) end)
  let incr () = set (1 + get ())
end

(** Build an integer option initialized to [0]. *)
module Zero(X: sig val name: string end) =
  Int(struct include X let default = 0 end)

(** {3 String} *)

module type STRING = S with type t = string

(** Build a string option. *)
module String(X: sig val default: string val name: string end) =
  Build(struct type t = string include X let equal = (=) end)

(** Build a string option initialized to [""]. *)
module EmptyString(X: sig val name: string end) =
  String(struct include X let default = "" end)

(** {3 String set} *)

module type STRING_SET = sig
  include S with type t = Cilutil.StringSet.t
  val add: string -> unit
  val remove: string -> unit
  val add_set: string -> unit
  val remove_set: string -> unit
end

(** Build an option as a set of strings, initialized to the empty set. *)
module StringSet(X: sig val name: string end) = struct
  open Cilutil
  include Build(struct
		  type t = StringSet.t
		  let default = StringSet.empty
		  let name = X.name
		  let equal = StringSet.equal
		end)
  let add x = set (StringSet.add x (get ()))
  let remove x = set (StringSet.remove x (get()))

  let split_set = Str.split (Str.regexp "[ \t]*,[ \t]*")
  let add_set x = set (List.fold_right StringSet.add (split_set x) (get ()))
  let remove_set x =
    set (List.fold_right StringSet.remove (split_set x) (get()))
end

(** {3 Complex values indexed by strings} *)

(** option interface *)
module type INDEXED_VAL = sig
  include STRING
  type value (** the real type for the option*)
  val add_choice: string -> value -> unit
    (** adds a new choice for the option. *)
  val get_val: unit -> value
end

(** input signature for [IndexedVal] *)
module type COMPLEX_VALUE = sig
  type t (** the type to be serialized *)
  val default_val: t (** the default value *)
  val default_key: string (** the default index *)
  val name: string (** name of the associated state *)
end

module IndexedVal (V:COMPLEX_VALUE):INDEXED_VAL with type value = V.t = struct

  type value = V.t

  let options = Hashtbl.create 13
  let add_choice k v  = Hashtbl.add options k v
  let () = add_choice V.default_key V.default_val

  let on = ref false

  let create () = ref V.default_key

  let curr_choice = ref (create())
  let cmdline_choice = ref (create())

  module StateAux =
  struct
    let name = Project.Computation.Name.make V.name
    let create = create

    type t = string ref

    let before_load () = ()
    let after_load () = ()
    let get () = !curr_choice
    let set s =
      if Hashtbl.mem options !s then begin
        curr_choice:=s; cmdline_choice:=s; on:=true
      end else
        Printf.eprintf
          "Warning: %s: identifier %s is not a valid index for this option. \
         Option is unchanged.\n" V.name !s

    let get_cmdline () = !cmdline_choice

    let dependencies = []
    let copy s = ref (!s)
    let rehash s = copy s
    let clear tbl = tbl:= V.default_key
  end

  module State =
    Project.Computation.Register
      (Datatype.Ref(Datatype.String))(StateAux)(StateAux)

  type t = string

  let () = add_option State.select

  include State

  let set s =
    if Hashtbl.mem options s then begin
      !curr_choice:=s; !cmdline_choice:=s; on:=true
    end else
      Printf.eprintf
        "Warning: %s: identifier %s is not a valid index for this option. \
         Option is unchanged.\n" V.name s

  let get () = !(!curr_choice)
  let get_val () = Hashtbl.find options (get())
  let get_cmdline () = !(!cmdline_choice)
  let clear () = !curr_choice:=V.default_key
  let is_set () = !on

end

(* ************************************************************************* *)
(** {2 Options} *)
(* ************************************************************************* *)

module Debug = Zero(struct let name = "Cmdline.debug" end)
module Quiet = False(struct let name = "Cmdline.quiet" end)
module Obfuscate = False(struct let name = "Cmdline.obfuscate" end)
module Metrics = False(struct let name = "Cmdline.metrics" end)
module ForceDeps = False(struct let name = "Cmdline.force deps" end)
module ForceCallDeps = False(struct let name = "Cmdline.force call deps" end)
module ForceUsers = False(struct let name = "Cmdline.force users" end)
module ForceMemzones = False(struct let name = "Cmdline.force memzones" end)
module HasForeignPointers =
  False(struct let name = "Cmdline.has foreign pointers" end)
module ForceValues = False(struct let name = "Cmdline.force values" end)
module ForceOut = False(struct let name = "Cmdline.force out" end)
module ForceInput = False(struct let name = "Cmdline.force input" end)
module ForceInout = False(struct let name = "Cmdline.force inout" end)
module ForceDeref = False(struct let name = "Cmdline.force deref" end)
module ForceAccessPath = False(struct let name = "Cmdline.force access path" end)
module UnsafeArrays = False(struct let name = "Cmdline.unsafe arrays" end)

module WideningLevel =
  Int(struct let default = 3 let name = "Cmdline.widening level" end)
module UnrollingLevel = Zero(struct let name = "Cmdline.unrolling level" end)
module SemanticUnrollingLevel =
  Zero(struct let name = "Cmdline.SemanticUnrollingLevel" end)
module ArrayPrecisionLevel =
  Int(struct let default = 200 let name = "Cmdline.array precision level" end)

module MainFunction =
  String(struct let default = "main" let name = "Cmdline.main function" end)
let () = MainFunction.depend Cil_state.self;

module LibEntry = EmptyString(struct let name = "Cmdline.lib entry" end)
let () = LibEntry.depend Cil_state.self

module PrintCode = False(struct let name = "Cmdline.print code" end)
module PrintComments = False(struct let name = "Cmdline.print comments" end)
module CodeOutput = EmptyString(struct let name = "Cmdline.code output" end)

module DoLoadState = False(struct let name = "Cmdline.do load state" end)
module DoSaveState = False(struct let name = "Cmdline.do save state" end)

module SaveState =
  String(struct let default = "framac.sav" let name = "Cmdline.SaveState" end)
module LoadState =
  String(struct let default = "framac.sav" let name = "Cmdline.LoadState" end)

module CallgraphFilename =
  EmptyString(struct let name = "Cmdline.callgraph filename" end)
module AlcoolPrintLocations =
  False(struct let name = "Cmdline.alcool print locations" end)
module AlcoolExtraction =
  EmptyString(struct let name = "Cmdline.alcool extraction" end)
module PointersExtraction =
  EmptyString(struct let name = "Cmdline.pointers extraction" end)
module XmlPointersExtraction =
  EmptyString(struct let name = "Cmdline.XML pointers extraction" end)
module RawExtraction =
  EmptyString(struct let name = "Cmdline.raw extraction" end)
module PromelaExtraction =
  EmptyString(struct let name = "Cmdline.promela extraction" end)
module Relevant = False(struct let name = "Cmdline.relevant" end)
module Report = EmptyString(struct let name = "Cmdline.report" end)
module ConstFuncArrays =
  False(struct let name = "Cmdline.Const func arrays" end)
module IgnoreOverflow = False(struct let name = "Cmdline.ignore overflow" end)
module MemFunctions = StringSet(struct let name = "Cmdline.mem functions" end)
module WidenVariables =
  StringSet(struct let name = "Cmdline.widen variables" end)
module ReadAnnot = True(struct let name = "read annotations" end)
module PreprocessAnnot =
  False(struct let name = "preprocess annotations" end)
module CppCommand =
  EmptyString(struct let name = "cpp command" end)
module CppExtraArgs =
  EmptyString(struct let name = "cpp extra arguments" end)
module WpCfg = False(struct let name = "Cmdline.wp cfg" end)
module MielSpecFilename =
  EmptyString(struct let name = "Cmdline.miel spec filename" end)
module PropagateTop = False(struct let name = "Cmdline.propagate top" end)
module LeafFuncReturnInt =
  False(struct let name = "Cmdline.leaf func return int" end)

module AutomaticContextMaxDepth =
  Int(struct
	let name = "Cmdline.AutomaticContextMaxDepth"
	let default = 2
      end)

module AllocatedContextValid =
  False(struct let name = "Cmdline.allocated contex valid" end)

module MemExecAll = False(struct let name = "Cmdline.mem exec all" end)

(** This can be set to false to debug value analysis without relations *)
module UseRelations = True(struct let name = "Cmdline.use relations" end)

module SimplifyCfg = False(struct let name = "Cmdline.simplify cfg" end)

module KeepSwitch = False(struct let name = "Cmdline.keep switch" end)

module KeepOnlyLastRun =
  False(struct let name = "Cmdline.keep only last run" end)

module MemoryFootprint =
  Int(struct let name = "Cmdline.memory footprint" let default = 2 end)

module FloatDigits =
  Int(struct
	let name = "Cmdline.float digits"
	let default = 4
      end)

module Files =
struct
  module M =
  Build(struct
	  type t = string list
	  let name = "Cmdline.files"
	  let default = []
	  let equal = (=)
	end)
  include M
  module Check = False (struct
                          let name = "Debug option: performs various \
                                      integrity checks on the AST"
                        end)
  module Copy = False (struct
                         let name = "Debug option: makes a copy of the initial \
                                     state before starting the real analyses"
                       end)
end


(** All absolute address are invalid *)
module MinValidAbsoluteAddress =
  Build(struct
	  type t = Abstract_interp.Int.t
	  let default = Abstract_interp.Int.one
	  let name = "Cmdline.min valid absolute address"
	  let equal = Abstract_interp.Int.equal
	end)

module MaxValidAbsoluteAddress =
  Build(struct
	  type t = Abstract_interp.Int.t
	  let default = Abstract_interp.Int.zero
	  let name = "Cmdline.max valid absolute address"
	  let equal = Abstract_interp.Int.equal
	end)

(** {3 Viewer options} *)

module MonospaceFontName =
  String(struct
	   let default = "Monospace 12"
	   let name = "Cmdline.MonospaceFontName"
	 end)

module GeneralFontName =
  String(struct
	   let default = "Helvetica 10"
	   let name = "Cmdline.GeneralFontName"
	 end)

(** {3 Security options} *)

module Security = struct

  module Analysis = False(struct let name = "Cmdline.security analysis" end)
  module Lattice =
    String(struct
	     let name = "Cmdline.security lattice"
	     let default = "weak"
	   end)
  module PropagateAssertions =
    False(struct let name = "Cmdline.propagate assertions" end)
  module Slicing = False(struct let name = "Cmdline.security slicing" end)

  let is_on () = Analysis.get_cmdline () || Slicing.get_cmdline ()

  module LogicAnnotation = struct
    include String(struct
		     let name = "Cmdline.Security.LogicAnnotation"
		     let default = "medium"
		   end)
    let get_cmdline () =
      let v = get_cmdline () in
      if is_on () && v = "" then Lattice.get_cmdline () else v
  end

end

(** {3 Jessie options} *)

module Jessie = struct
  module ProjectName = EmptyString(struct let name = "Cmdline.jessie project name" end)
  module Analysis = False(struct let name = "Cmdline.jessie analysis" end)
  module Gui = False(struct let name = "Cmdline.jessie gui" end)
  module WhyOpt = StringSet(struct let name = "Cmdline.jessie why opt" end)
end

(** {3 PDG options} *)

module Pdg = struct
  module BuildAll = False(struct let name = "Cmdline.build pdg" end)
  module BuildFct = StringSet(struct let name = "Cmdline.build fct pdg" end)
  module DotBasename =
    EmptyString(struct let name = "Cmdline.pdg dot basename" end)
  module DotPostdomBasename =
    EmptyString(struct let name = "Cmdline.pdg postdominators dot basename" end)
  module Verbosity =
    Int (struct let name = "Pdg verbosity level" let default = 0 end)
end

(** {2 Sparecode options} *)

module Sparecode = struct
  module Analysis = False(struct let name = "Cmdline.sparecode analysis" end)
end


(** {3 Slicing options} *)

module Slicing = struct
  module Select = struct
    module Calls =  StringSet(struct let name = "Cmdline.slice calls" end)
    module Return = StringSet(struct let name = "Cmdline.slice return" end)
    module Threat = StringSet(struct let name = "Cmdline.slice threat" end)
    module Assert = StringSet(struct let name = "Cmdline.slice assert" end)
    module LoopInv = StringSet(struct let name = "Cmdline.slice loop invariant" end)
    module LoopVar = StringSet(struct let name = "Cmdline.slice loop variant" end)
    module Pragma = StringSet(struct let name = "Cmdline.slice pragma" end)
    module RdAccess = StringSet(struct let name = "Cmdline.slice read access" end)
    module WrAccess = StringSet(struct let name = "Cmdline.slice write access" end)
    module Value = StringSet(struct let name = "Cmdline.slice value" end)
  end
  module Mode = struct
    module Verbose = Zero(struct let name = "Cmdline.slice verbosity" end)
    module Callers = True(struct let name = "Cmdline.slice callers" end)
    module Calls = Int(struct let name = "Cmdline.how to slice calls" let default = 2 end)
    module SliceUndef = False(struct let name = "Cmdline.slice undefined functions" end)
  end
  module Print = False(struct let name = "Cmdline.print sliced code" end)
end

(** {3 Constant propagation options} *)

module Constfold = False(struct let name = "Cmdline.constfold" end)
module Constant_Propagation = struct
  module SemanticConstFolding =
    False(struct let name = "Cmdline.SemanticConstFolding" end)
  module SemanticConstFold =
    StringSet(struct let name = "Cmdline.SemanticConstFold" end)
end

(** {3 Cxx options} *)

module Unmangling: INDEXED_VAL with type value = string -> string =
  IndexedVal(
    struct
      type t = string->string
      let default_val = fun x -> x
      let default_key = "id"
      let name = "Cmdline.unmangling option"
    end)

module PrintCxx = False(struct let name = "pretty-print cxx files" end)


(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.."
End:
*)
