(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007                                                    *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* $Id: loop.ml,v 1.9 2008/03/06 14:10:17 uid528 Exp $ *)

open Cil_types
open Db_types
open Cilutil
open Cil

module Natural_Loops =
  Kernel_function.Make_Table
    (Project.Datatype.Imperative
       (struct
	  type t = (stmt * stmt list) list
	  let copy _ = assert false (* TODO: deep copy *)
	end))
    (struct
       let name = Project.Computation.Name.make "Natural_Loops"
       let size = 97
       let dependencies = [ Cil_state.self ]
	 (* [TODO JS 2007/10/05:] change it! *)
     end)

let rec pretty_natural_loops fmt kf =
  Format.fprintf fmt "Natural_loops for %s:@." (Kernel_function.get_name kf);
  let loops = get_naturals kf in
  List.iter
    (fun (start,members) ->
       Format.fprintf fmt "Loop start: %d ( " start.sid;
       List.iter (fun d -> Format.fprintf fmt "%d " d.sid) members;
       Format.fprintf fmt ")@\n";)
    loops

and get_naturals kf =
  let memo = ref false in
  let loops =
    Natural_Loops.memo
      (fun kf ->
	 memo := true;
	 match kf.fundec with
	 | Declaration _ ->
	     []
	 | Definition (cilfundec,_) ->
	     if Cmdline.Debug.get () > 0 || Cmdline.Pdg.Verbosity.get() > 0
             then
	       Format.printf "COMPUTE NATURAL LOOPS FOR %S@."
		 (Kernel_function.get_name kf);
	     let dominators = Dominators.computeIDom cilfundec in
	     let naturals = Dominators.findNaturalLoops cilfundec dominators in
	     naturals)
      kf
  in
  if (Cmdline.Debug.get() > 0 || Cmdline.Pdg.Verbosity.get() > 0) && !memo then
   begin
    Format.printf "DONE COMPUTE NATURAL LOOPS FOR %S@."
      (Kernel_function.get_name kf);
    pretty_natural_loops Format.std_formatter kf;
  end;
  loops

let is_natural kf =
  let natural_loops =
    List.fold_left
      (fun acc (n,_) -> StmtSet.add n acc)
      StmtSet.empty
      (get_naturals kf)
  in
(* non natural loop over-approximation try:
  let can_reach = !stmt_can_reach kf in *)
  fun stmt -> let nat_loop = StmtSet.mem stmt natural_loops in
  nat_loop
(*  if nat_loop then nat_loop
  else
    if can_reach stmt stmt
    then true (* this is non natural loop or an imbricated loop... *)
    else false
*)
let back_edges kf stmt =
  if is_natural kf stmt then
    let rec lookup = function
      | [] -> assert false
      | (s, pred_s) :: sl -> if s.sid = stmt.sid then pred_s else lookup sl
    in
    lookup (get_naturals kf)
  else
    []

let while_for_natural_loop kf stmt =
  match stmt.skind with
  | Loop _ -> stmt
  | _ -> (* the while stmt is probably the non looping predecessor *)
      let be = back_edges kf stmt in
      Format.printf "Stmt:%d " stmt.sid;
      List.iter (fun x -> Format.printf "B_edge:%d " x.sid) be;
      List.iter (fun x -> Format.printf "Preds:%d " x.sid) stmt.preds;
      let non_looping_pred =
        List.filter (fun pred -> not (List.mem pred be)) stmt.preds in
      match non_looping_pred with
      | [x] -> x
      | _ ->
          Format.eprintf "@.Lexical non natural loop detected !@.";
          assert false

let compute_allstmt_block block =
  let visitor = object
    val mutable allstmts = StmtSet.empty
    method allstmts = allstmts
    inherit nopCilVisitor as super
    method vstmt s = allstmts <- StmtSet.add s allstmts;
      DoChildren
  end
  in
  ignore (visitCilBlock (visitor:>cilVisitor) block);
  visitor#allstmts

let compute_loops_stmts kf =
  let tbl = InstrHashtbl.create 17 in
  let visitor = object
    inherit nopCilVisitor as super
    method vstmt s =
      (match s.skind with
       | Loop (_,block,_,_,_) ->
           InstrHashtbl.add tbl (Kstmt s) (compute_allstmt_block block)
       |  _ -> ());
      DoChildren
  end
  in
  (try
     ignore
       (visitCilFunction
	  (visitor :> cilVisitor)
	  (Kernel_function.get_definition kf));
   with Kernel_function.No_Definition -> ());
  tbl

exception No_such_while

(** @raise No_such_while if [stmt.skind] is not a [While]. *)
let get_loop_stmts =
  let module S =
    Kernel_function.Make_Table
      (Kernel_datatype.InstrHashtbl(Kernel_datatype.StmtSet))
      (struct
	 let name = Project.Computation.Name.make "LoopStmts"
	 let size = 97
	 let dependencies = [ Cil_state.self ] (* JS 2007/08/30: is it ok? *)
       end)
  in
  fun kf loop_stmt ->
    (match loop_stmt.skind with
     | Loop _ -> ()
     | _ -> raise No_such_while);
    let tbl = S.memo compute_loops_stmts kf in
    try InstrHashtbl.find tbl (Kstmt loop_stmt) with Not_found -> assert false

(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.."
End:
*)
