(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007                                                    *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(* $Id: visitor.mli,v 1.11 2008/03/06 14:10:17 uid528 Exp $ *)

open Cil_types
open Db_types

(** Frama-C visitors dealing with projects. *)

(** Class type for a Db-aware visitor.
    This is done by defining auxiliary methods that can be
    redefined in inherited classes, while the corresponding ones from
    {!Cil.cilVisitor} {b must} retain their values as defined here. Otherwise,
    annotations may not be visited properly. The replaced functions are
    - {t vfile} (use {t vfile_aux} instead)
    - {t vstmt} (use {t vstmt_aux} instead)
    - {t vglob} (use {t vglob_aux} instead)

    {b A few hints on how to use correctly this visitor}

    - when initializing a new project with it
    (see {!File.init_project_from_visitor}), use a visitor with copy behavior

    - [SkipChildren] and [ChangeTo] must be used with extreme care in a visitor
    with copy behavior, or some nodes may be shared between the original and
    the copy.

    - Do not erase a statement during the visit, as there might be
    annotations attached to it. Change it to Skip instead, the
    [generic_frama_c_visitor] will know what to do.

    - Be careful if you change the [vid] or [sid]: this must be done before
    anything has been attached to the corresponding variable or
    statement in the new project, which means
       - for statements, in [vstmt], for the current statement only
       - for variables, at their declaration point.
 *)
class type frama_c_visitor = object
  inherit Cil.cilVisitor
  method vfile_aux: file -> file Cil.visitAction
    (** replacement of vfile. *)
  method vstmt_aux: stmt -> stmt Cil.visitAction
    (** Replacement of vstmt. *)
  method vglob_aux: global -> global list Cil.visitAction
    (** Replacement of vglob. *)
  method vrooted_code_annotation:
    rooted_code_annotation ->
    rooted_code_annotation list Cil.visitAction
      (** visiting a rooted code annotation. *)
  method is_annot_before: bool
    (** Used to tell if we're visiting an annotation placed
	before current statement.
	@raise Error if not called while visiting a statement. *)
  method fill_global_tables: unit
    (** due to efficiency reasons, the global tables related to GFun, GVar, ...
        are not filled after each visit of a global, but at the end of the
        visit of a file. When visiting a single global, outside of a file,
        it might be needed to call this method explicitely. *)
  method get_filling_actions: (unit -> unit) Queue.t
    (** gets the pending actions that must be performed to fill global tables.*)
end

class generic_frama_c_visitor:
  Cil.visitor_behavior -> Project.t -> frama_c_visitor
  (** generic visitor. The tables are filled on the project given in argument.
      See also {!File.init_project_from_visitor}. *)

class frama_c_copy: Project.t -> frama_c_visitor
  (** copying visitor *)

(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.."
End:
*)
