(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007                                                    *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Locations
open Abstract_interp
open Abstract_value

(* To be raised whenever we need to fall back to value analysis *)
exception Use_Main_Memory

module type Model_S = sig 
  type t
  type widen_hint = Cvalue_type.Model.widen_hint
  type cluster
  val rehash: t -> t
  val is_reachable : t -> bool
  val pretty : Format.formatter -> t -> unit
  val pretty_without_null : Format.formatter -> t -> unit
  val pretty_filter : Format.formatter -> t -> Zone.t -> unit
  val join : t -> t -> t

  val find : with_alarms:CilE.warn_mode -> t -> location -> Location_Bytes.t
  val find_unspecified : with_alarms:CilE.warn_mode -> t -> location -> 
    Cvalue_type.V_Or_Uninitialized.t

  val add_binding : 
    with_alarms:CilE.warn_mode -> exact:bool -> t -> location -> Location_Bytes.t -> t
  val add_binding_unspecified : t -> location -> t

  val reduce_binding : t -> location -> Location_Bytes.t -> t
  val is_included : t -> t -> bool
  val equal : t -> t -> bool
  val is_included_actual_generic : 
    Zone.t -> t -> t -> Locations.Location_Bytes.t BaseUtils.BaseMap.t
  val widen : widen_hint -> t -> t -> (bool * t)
  val bottom : t
  val inject : Cvalue_type.Model.t -> t
  val empty : t
  val value_state : t -> Cvalue_type.Model.t

  (** May raise [Lmap.Cannot_copy]. *)
  val paste_offsetmap : 
    Cvalue_type.V_Offsetmap.t -> Location_Bits.t -> Int.t -> Int.t -> t -> t
    
  (** May raise [Lmap.Cannot_copy]. *)
  val copy_paste : location  -> location -> t -> t
    
  (** May raise [Lmap.Cannot_copy]. *)
  val copy_from_virtual : 
    location ->
    Ival.t ->
    Int.t -> t -> Cvalue_type.V_Offsetmap.t

  (** May raise [Lmap.Cannot_copy]. *)  
  val copy_offsetmap : Locations.location -> t -> Cvalue_type.V_Offsetmap.t option
  val drop_relations : t -> t
  val filter_base : (Base.t -> bool) -> t -> t 
  val clear_state_from_locals : Cil_types.fundec -> t -> t

  val compute_actual_final_from_generic :
    t -> t -> Zone.t -> Cvalue_type.Model.instanciation -> t
  val is_included_by_location_enum :  t -> t -> Zone.t -> bool

  (* computes the value of [*location] *)
  val find_mem : location -> Int_Base.t ->
    Ival.t -> t -> Cvalue_type.V.t
  (* add an information about the value of [*location] *)
  val add_mem : location -> Int_Base.t -> Ival.t -> t -> Cvalue_type.V.t -> cluster list*t
  (* clean up relation about [location] (call this function each time location has changed) *)
  val propagate_change_from_real_to_virt : 
    protected_clusters:cluster list -> location -> t -> Cvalue_type.V.t -> t
  val add_equality : ?offset:Ival.t -> t -> location -> location -> t
  val reduce_equality : t -> location -> location -> t    
  val compute_diff : t -> location -> location -> Cvalue_type.V.t

  val shift_location : t -> location -> Ival.t -> Cvalue_type.V.t -> t

  (** Raise [Not_found] when the vid is not in the map *)
  val find_base : Base.t -> t -> Cvalue_type.V_Offsetmap.t

  val create_initial : base:Base.t ->
    size:int ->
    v:Cvalue_type.V.t ->
    modu:Int.t ->
    state:t -> t
(* [size] is in bits. For instance [~size:40] for creating an 
   initial variable on a 32-bit architecture. *)
    
  module Datatype : Project.Datatype.OUTPUT with type t = t
end

module Model : Model_S
(*
Local Variables:
compile-command: "make -C ../.."
End:
*)
