(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007                                                    *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Format

type section = { fullname: string;
                 mutable cmdline: (Arg.key * Arg.spec * Arg.doc) list ;
                 shortname: string;
                 mutable debug: (Arg.key * Arg.spec * Arg.doc) list }

type plugin = { name: string; descr: string; config: unit -> unit }

let available_plugins : plugin list ref = ref []
let available_sections : section list ref = ref []

let has_plugin name = List.exists (fun x -> x.name = name) !available_plugins
let has_section name =
  List.exists (fun x -> x.shortname = name) !available_sections

let find_section ~shortname =
  assert (has_section shortname);
  List.find (fun x -> x.shortname = shortname) !available_sections

let usage () =
  let usage = "toplevel options files..." in
  let plugins =
    match !available_plugins with
    | [] -> usage
    | l ->
        List.fold_left
          (fun s plg ->
             Cil.fprintf_to_string "@[%s@\n  @[%s:@ %a@]@]"
               s plg.name Pretty_utils.pp_print_string_fill plg.descr)
          (usage ^ "\n\nPLUGINS BUILT:") l
  in
  plugins ^ "\n\nAVAILABLE OPTIONS:"

let add_file f = Cmdline.Files.set (f :: Cmdline.Files.get ())

let section name =
  "", Arg.Unit (fun () -> assert false), "\n*** " ^ String.uppercase name

(* The list of global options *)
let cmdline_to_parse = ref []

let add_cmdline x = cmdline_to_parse := !cmdline_to_parse @ x

let add_to_cmdline_section ~shortname ~debug options =
  let section = find_section shortname in
  section.cmdline <- section.cmdline @ options ;
  section.debug <- section.debug @ debug

let add_cmdline_section ~name ~shortname ~debug options =
  let section = { fullname = String.capitalize name ; cmdline = options ;
                  shortname = String.uncapitalize shortname ; debug = debug}
  in available_sections := section :: !available_sections

let add_cmdline ?name ?shortname ?(debug=[]) options =
  match name with
  | None ->
      begin
        match shortname with
        | None -> assert (debug = []); add_cmdline options
        | Some shortname -> add_to_cmdline_section ~shortname ~debug options
      end
  | Some name ->
      begin
        match shortname with
        | None -> add_cmdline_section ~name ~shortname:name ~debug options
        | Some shortname -> add_cmdline_section ~name ~shortname ~debug options
      end

let () = add_cmdline
    [section "general options" ;

     "-version",
     Arg.Unit
       (fun () -> Format.printf "Version: %s@\nCompilation date: %s@\nFrama-C library path: %s (may be overridden with FRAMAC_SHARE variable)@."
	  Version.version Version.date Version.dataroot),
     ": print version information";

    "-ocode",
    Arg.String Cmdline.CodeOutput.set,
    "filename : when printing code, redirects the output to file [filename].";

     "-save",
     Arg.String
       (fun s -> Cmdline.DoSaveState.on (); Cmdline.SaveState.set s),
     "filename : save the state into file [filename] after computations.";
     "-load",
     Arg.String
       (fun s -> Cmdline.DoLoadState.on (); Cmdline.LoadState.set s),
     "filename : load an initial (previously saved) state from file [filename].";
     "-time",
     Arg.String
       (fun s ->
          at_exit (fun () ->
                     let oc = open_out_gen
                       [Open_append;Open_creat;Open_binary]
                       0b111100100
                       s
                     in
                     let {Unix.tms_utime = time } = Unix.times () in
                     let now = Unix.localtime (Unix.time ()) in
                       Printf.fprintf oc "%02d/%02d/%02d %02d:%02d:%02d %f\n"
                         now.Unix.tm_mday
                         (now.Unix.tm_mon+1)
                         (now.Unix.tm_year - 100)
                         now.Unix.tm_hour
                         now.Unix.tm_min
                         now.Unix.tm_sec
                         time;
                       flush oc;
                       close_out oc)
       ),
     "filename : append user time and date to [filename] at exit.";

     "-quiet",
     Arg.Unit Cmdline.Quiet.on,
     ": do not print results of analyzes on stdout.";

     "-debug", Arg.Int Cmdline.Debug.set,
     Format.sprintf "n : level of debug (defaults to %d)."
       (Cmdline.Debug.get ());

     section "syntactical tools" ;

     "-print",
     Arg.Unit Cmdline.PrintCode.on,
     " : pretty print original code with its comments.";

     "-simplify-cfg",
     Arg.Unit Cmdline.SimplifyCfg.on,
     ": remove break, continue and switch statement before analyzes.";

     "-keep-switch",
     Arg.Unit Cmdline.KeepSwitch.on,
     ": keep switch statements despite -simplify-cfg.";

     "-keep-comments",
     Arg.Unit Cmdline.PrintComments.on,
     Format.sprintf
       ": try to keep comments in C code (defaults to %b)."
       (Cmdline.PrintComments.get ());

     "-ulevel",
     Arg.Int Cmdline.UnrollingLevel.set,
     Format.sprintf "n : unroll loops n times (defaults to %d) before analyzes."
       (Cmdline.UnrollingLevel.get ());

     "-constfold",
     Arg.Unit Cmdline.Constfold.on,
     ": fold all constant expressions in the code before analysis.";

     "-obfuscate",
     Arg.Unit Cmdline.Obfuscate.on,
     ": print an obfuscated version of files to standard output and exit.";

     "-metrics",
     Arg.Unit Cmdline.Metrics.on,
     ": print some metrics.";

     "-machdep",
     Arg.String (fun s -> match s with
                 | "x86_16" ->
                     let module M = Machdep.DEFINE(Machdep_x86_16) in
                     ();
                 | "x86_32" ->
                     let module M = Machdep.DEFINE(Machdep_x86_32) in
                     ();
                 | "x86_64" ->
                     let module M = Machdep.DEFINE(Machdep_x86_64) in
                     ();
                 | "ppc_32" ->
                     let module M = Machdep.DEFINE(Machdep_ppc_32) in
                     ();
                 | "ppc_32_diab" ->
                     let module M = Machdep.DEFINE(Machdep_ppc_32_diab) in
                     ();
                 | _ -> Format.printf "Unsupported machine %s@." s;
                     exit 1
                ),
     "machine : use [machine] as the current machine dependent configuration.";

     "-msvc",
     Arg.Set Cil.msvcMode,
     ": switch to MSVC mode. Default mode is gcc."

    ]

module Startup_Hook = Hook.Make(struct end)
let initialize_toplevels = Startup_Hook.apply

module Init_Hook = Hook.Make(struct end)
let initialize_plugins = Init_Hook.apply
let register_plugin_init = Init_Hook.extend

let () =
  Init_Hook.extend (fun () -> Project.set_current (Project.create "default"))

let add_plugin ~name ~descr
    ?plugin_init ?(init=fun () -> ()) ?toplevel_init ?(shortname=name)
    ?(debug=[])
    options =
  let plugin =
    { name = String.capitalize name; descr = descr; config = init }
  in
  available_plugins := plugin :: !available_plugins;
  add_cmdline_section ~name:plugin.name ~shortname ~debug options;
  begin match toplevel_init with
    | None -> ()
    | Some f -> Startup_Hook.extend f
  end;
  match plugin_init with
  | None -> ()
  | Some f -> Init_Hook.extend f

let build_debug_options plugin options =
  let name = "-"^plugin^"-debug" in
  let parse_debug s =
    let argv = Array.of_list (Sys.argv.(0)::(Str.split (Str.regexp "[ \t;]+") s)) in
    let current = ref 0 in
      (try
        Arg.parse_argv ~current argv options
          (fun s -> raise (Arg.Bad (Format.sprintf "Invalid argument to %S:'%S'" name s)))
          ("This is the list of internal options of "^name^".")
      with
        | Arg.Bad mesg -> Format.eprintf "%s@." mesg;exit 2
        | Arg.Help mesg -> Format.eprintf "%s@." mesg;exit 0)
  in name, Arg.String parse_debug, ("sub-options: use '"^name^" -help' to get information about internal sub-options of "^plugin)

let parse_cmdline () =
  let cmdline =  List.fold_left
    (fun s sect ->
       match sect.cmdline, sect.debug with
         | [],[] -> s
         | cmdline,[] -> section sect.fullname :: cmdline @ s
         | cmdline,debug ->
             let debug = build_debug_options sect.shortname debug
             in section sect.fullname :: cmdline @ ( debug :: s))
    [ section "HELP" ] !available_sections

  in
  Arg.parse (!cmdline_to_parse @ cmdline) add_file (usage());
  List.iter
    (fun x ->
       try
         x.config ();
         if Cmdline.Debug.get () > 0 then
           Printf.eprintf "%s registered\n%!" x.name;
       with exn ->
         if Cmdline.Debug.get () > 0 then
           Printf.eprintf
             "Warning, configuration of %s failed:\n%s\n%!"
             x.name (Printexc.to_string exn))
    !available_plugins


(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.."
End:
*)
